@@ -1,5 +1,5 @@
 #!/usr/bin/python
-# MySqlPy V1.5.0
+# MySqlPy V1.5.1
 # Author: Luca.Canali@cern.ch
 # 
 #
@@ -9,7 +9,7 @@
 #           http://catherine.devlin.googlepages.com/
 
 from sqlpyPlus import *
-import binascii, sys, tempfile
+import sys, tempfile, optparse, unittest
 
 class mysqlpy(sqlpyPlus):
     '''
@@ -104,11 +104,7 @@ WHERE  created > SYSDATE - 7;''')
         '''Runs query_load10g defined above, to display OS load on cluster nodes (10gRAC)
 Do not confuse with `GET myfile.sql` and `@myfile.sql`,
 which get and run SQL scripts from disk.'''
-        self.do_select(self.query_load10g)
-
-    def do_himom(self,args):
-        '''greets your mom'''
-        print 'hi mom'
+        self.onecmd(self.query_load10g)
 
     def do_db(self,args,filepath='pass.txt'): 
         '''Exec do_connect to db_alias in args (credentials form the file pass.txt) '''
@@ -125,9 +121,8 @@ which get and run SQL scripts from disk.'''
             print e
 
     def do_tselect(self, arg):  
-        '''executes a query and prints the result in trasposed form. Useful when querying tables with many columns''' 
-        
-        self.do_select(arg, override_terminator='\\t')            
+        '''executes a query and prints the result in trasposed form. Useful when querying tables with many columns'''  
+        self.do_select(self.parsed(arg, terminator='\\t'))
 
     def do_sql(self,args):
         '''prints sql statement give the sql_id (Oracle 10gR2)'''
@@ -157,8 +152,17 @@ which get and run SQL scripts from disk.'''
             print e
 
     def do_sessinfo(self,args):
-        '''Reports session info for the give sid, extended to RAC with gv$'''
-        self.do_tselect('* from gv$session where sid='+args+';')
+        '''Reports session info for the given sid, extended to RAC with gv$'''
+        try:
+            if not args:
+                self.curs.execute('SELECT sid FROM v$mystat')
+                args = self.curs.fetchone()[0]
+            self.onecmd('SELECT * from gv$session where sid=%s\\t' % args)
+        except cx_Oracle.DatabaseError, e:
+            if 'table or view does not exist' in str(e):
+                print 'This account has not been granted SELECT privileges to v$mystat or gv$session.'
+            else:
+                raise 
 
     def do_sleect(self,args):    
         '''implements sleect = select, a common typo'''
@@ -188,5 +192,16 @@ def run():
         pass
     my.cmdloop()
 
+class TestCase(Cmd2TestCase):
+    CmdApp = mysqlpy
+    transcriptFileName = 'exampleSession.txt'
+
 if __name__ == '__main__':
-    run()        
+    parser = optparse.OptionParser()
+    parser.add_option('-t', '--test', dest='unittests', action='store_true', default=False, help='Run unit test suite')
+    (callopts, callargs) = parser.parse_args()
+    if callopts.unittests:
+        sys.argv = [sys.argv[0]]  # the --test argument upsets unittest.main()
+        unittest.main()
+    else:
+        run()
\ No newline at end of file
