@@ -2,7 +2,7 @@
 
 Features include:
  - SQL*Plus-style bind variables
- - Query result stored in special bind variable ":_" if one row, one item
+ - `set autobind on` stores single-line result sets in bind variables automatically
  - SQL buffer with list, run, ed, get, etc.; unlike SQL*Plus, buffer stores session's full history
  - @script.sql loads and runs (like SQL*Plus)
  - ! runs operating-system command
@@ -23,18 +23,18 @@ or with a python-like shorthand
 
 - catherinedevlin.blogspot.com  May 31, 2006
 """
-import sys, os, re, sqlpython, cx_Oracle, pyparsing, re, completion, datetime, pickle
-from cmd2 import Cmd, make_option, options, Statekeeper
+import sys, os, re, sqlpython, cx_Oracle, pyparsing, re, completion, datetime, pickle, binascii
+from cmd2 import Cmd, make_option, options, Statekeeper, Cmd2TestCase
 from output_templates import output_templates
 from plothandler import Plot
 try:
     import pylab
-except:
+except (RuntimeError, ImportError):
     pass
 
 descQueries = {
 'TABLE': ("""
-          atc.column_name,
+SELECT atc.column_name,
 CASE atc.nullable WHEN 'Y' THEN 'NULL' ELSE 'NOT NULL' END "Null?",
 atc.data_type ||
 CASE atc.data_type WHEN 'DATE' THEN ''
@@ -51,7 +51,7 @@ WHERE atc.table_name = :object_name
 AND      atc.owner = :owner
 ORDER BY atc.column_id;""",),
 'PROCEDURE': ("""
-              NVL(argument_name, 'Return Value') argument_name,             
+SELECT NVL(argument_name, 'Return Value') argument_name,             
 data_type,
 in_out,
 default_value
@@ -66,7 +66,7 @@ FROM all_arguments
 WHERE package_name = :package_name
 AND      owner = :owner""",),
 'PackageObjArgs':("""
-                  object_name,
+SELECT object_name,
 argument_name,             
 data_type,
 in_out,
@@ -76,15 +76,15 @@ WHERE package_name = :package_name
 AND      object_name = :object_name
 AND      owner = :owner
 AND      argument_name IS NOT NULL
-ORDER BY sequence""",),
+ORDER BY sequence;""",),
 'TRIGGER':("""
-           description
+SELECT description
 FROM   all_triggers
 WHERE  owner = :owner
-AND    trigger_name = :object_name
+AND    trigger_name = :object_name;
 """,
 """
-table_owner,
+SELECT table_owner,
 base_object_type,
 table_name,
 column_name,
@@ -99,7 +99,7 @@ AND    trigger_name = :object_name
 """,
 ),
 'INDEX':("""
-index_type,
+SELECT index_type,
 table_owner,
 table_name,
 table_type,
@@ -156,6 +156,7 @@ FROM    all_tab_comments
 WHERE owner = :owner
 AND      table_name = :table_name""",
 'colComments': """
+SELECT
 atc.column_name,
 acc.comments             
 FROM all_tab_columns atc
@@ -163,6 +164,15 @@ JOIN all_col_comments acc ON (atc.owner = acc.owner and atc.table_name = acc.tab
 WHERE atc.table_name = :object_name
 AND      atc.owner = :owner
 ORDER BY atc.column_id;""",
+'oneColComments': """
+SELECTatc.column_name,
+acc.comments             
+FROM all_tab_columns atc
+JOIN all_col_comments acc ON (atc.owner = acc.owner and atc.table_name = acc.table_name and atc.column_name = acc.column_name)
+WHERE atc.table_name = :object_name
+AND      atc.owner = :owner
+AND      acc.column_name = :column_name
+ORDER BY atc.column_id;""",
 #thanks to Senora.pm for "refs"
 'refs': """
 NULL               referenced_by, 
@@ -340,26 +350,27 @@ def findBinds(target, existingBinds, givenBindVars = {}):
             if not givenBindVars.has_key(varname):
                 print 'Bind variable %s not defined.' % (varname)                
     return result
-       
+
 class sqlpyPlus(sqlpython.sqlpython):
     defaultExtension = 'sql'
-    sqlpython.sqlpython.shortcuts.update({':': 'setbind', '\\': 'psql', '@': '_load'})
+    sqlpython.sqlpython.shortcuts.update({':': 'setbind', 
+                                          '\\': 'psql', 
+                                          '@': '_load'})
     multilineCommands = '''select insert update delete tselect
-                      create drop alter'''.split()
+                      create drop alter _multiline_comment'''.split()
+    sqlpython.sqlpython.noSpecialParse.append('spool')
+    commentGrammars = pyparsing.Or([pyparsing.Literal('--') + pyparsing.restOfLine, pyparsing.cStyleComment])
     defaultFileName = 'afiedt.buf'
     def __init__(self):
         sqlpython.sqlpython.__init__(self)
         self.binds = CaselessDict()
-        self.sqlBuffer = []
-        self.settable = ['maxtselctrows', 'maxfetch', 'autobind', 
-                         'failover', 'timeout', 'commit_on_exit'] # settables must be lowercase
+        self.settable = 'autobind commit_on_exit echo maxfetch maxtselctrows timeout'.split()
+        # settables must be lowercase
         self.stdoutBeforeSpool = sys.stdout
         self.spoolFile = None
         self.autobind = False
-        self.failover = False
-    def default(self, arg, do_everywhere=False):
-        sqlpython.sqlpython.default(self, arg, do_everywhere)
-        self.sqlBuffer.append(self.query)            
+    #def default(self, arg):
+    #    sqlpython.sqlpython.default(self, arg)
 
     # overrides cmd's parseline
     def parseline(self, line):
@@ -398,35 +409,20 @@ class sqlpyPlus(sqlpython.sqlpython):
         if needsquotes:
             return "'%s'" % str(itm)
         return str(itm)
-    def str_or_empty(self, itm):
-        if itm is None:
-            return ''
-        return str(itm)
-    def output_as_insert_statements(self):
-        usequotes = [d[1] != cx_Oracle.NUMBER for d in self.curs.description]
-        def formatRow(row):
-            return ','.join(self.sql_format_itm(itm, useq)
-                            for (itm, useq) in zip(row, usequotes))
-        result = ['INSERT INTO %s (%s) VALUES (%s);' %
-                  (self.tblname, ','.join(self.colnames), formatRow(row))
-                  for row in self.rows]
-        return '\n'.join(result)
     tableNameFinder = re.compile(r'from\s+([\w$#_"]+)', re.IGNORECASE | re.MULTILINE | re.DOTALL)          
+    inputStatementFormatters = {
+        cx_Oracle.STRING: "'%s'",
+        cx_Oracle.DATETIME: "TO_DATE('%s', 'YYYY-MM-DD HH24:MI:SS')"}
+    inputStatementFormatters[cx_Oracle.CLOB] = inputStatementFormatters[cx_Oracle.STRING]
+    inputStatementFormatters[cx_Oracle.TIMESTAMP] = inputStatementFormatters[cx_Oracle.DATETIME]    
     def output(self, outformat, rowlimit):
         self.tblname = self.tableNameFinder.search(self.curs.statement).group(1)
         self.colnames = [d[0] for d in self.curs.description]
-        if outformat == '\\i':
-            result = self.output_as_insert_statements()
-        elif outformat in output_templates:
+        if outformat in output_templates:
             self.colnamelen = max(len(colname) for colname in self.colnames)
-            result = output_templates[outformat].generate(**self.__dict__)
-        elif outformat in ('\\s', '\\S', '\\c', '\\C'): #csv
-            result = []
-            if outformat in ('\\s', '\\c'):
-                result.append(','.join('"%s"' % colname for colname in self.colnames))
-            for row in self.rows:
-                result.append(','.join('"%s"' % self.str_or_empty(itm) for itm in row))
-            result = '\n'.join(result)
+            self.coltypes = [d[1] for d in self.curs.description]
+            self.formatters = [self.inputStatementFormatters.get(typ, '%s') for typ in self.coltypes]    
+            result = output_templates[outformat].generate(**self.__dict__)        
         elif outformat == '\\t': # transposed
             rows = [self.colnames]
             rows.extend(list(self.rows))
@@ -499,14 +495,13 @@ class sqlpyPlus(sqlpython.sqlpython):
         return completions
     
     rowlimitPattern = pyparsing.Word(pyparsing.nums)('rowlimit')
-    rawTerminators = '; \\s \\S \\c \\C \\t \\i \\p \\l \\L \\b ' + ' '.join(output_templates.keys())
-    terminatorPattern = (pyparsing.oneOf(rawTerminators)    
-                        ^ pyparsing.Literal('\n/') ^ \
-                        (pyparsing.Literal('\nEOF') + pyparsing.stringEnd)) \
-                        ('terminator') + \
-                        pyparsing.Optional(rowlimitPattern) #+ \
-                        #pyparsing.FollowedBy(pyparsing.LineEnd())
-    def do_select(self, arg, bindVarsIn=None, override_terminator=None):
+    terminators = '; \\C \\t \\i \\p \\l \\L \\b '.split() + output_templates.keys()
+
+    def expandSelect(self, arg):
+        #w + Optional(CaselessKeyword('AS')) + Optional(w ^ dblQuotedString) + (CaselessKeyword('FROM') ^ ',')
+        return arg
+    
+    def do_select(self, arg, bindVarsIn=None, terminator=None):
         """Fetch rows from a table.
 
         Limit the number of rows retrieved by appending
@@ -517,44 +512,44 @@ class sqlpyPlus(sqlpython.sqlpython):
         ("help terminators" for details)
         """
         bindVarsIn = bindVarsIn or {}
-        statement = self.parsed('select ' + arg)
-        self.query = statement.unterminated
-        if override_terminator:
-            statement['terminator'] = override_terminator
-        statement['rowlimit'] = int(statement.rowlimit or 0)
         try:
-            self.varsUsed = findBinds(self.query, self.binds, bindVarsIn)
-            self.curs.execute(self.query, self.varsUsed)
-            self.rows = self.curs.fetchmany(min(self.maxfetch, (statement.rowlimit or self.maxfetch)))
-            self.desc = self.curs.description
-            self.rc = self.curs.rowcount
-            if self.rc > 0:
-                self.stdout.write('\n%s\n' % (self.output(statement.terminator, statement.rowlimit)))
-            if self.rc == 0:
-                print '\nNo rows Selected.\n'
-            elif self.rc == 1: 
-                print '\n1 row selected.\n'
-                if self.autobind:
-                    self.binds.update(dict(zip([''.join(l for l in d[0] if l.isalnum()) for d in self.desc], self.rows[0])))
-                    if len(self.desc) == 1:
-                        self.binds['_'] = self.rows[0][0]
-            elif self.rc < self.maxfetch:
-                print '\n%d rows selected.\n' % self.rc
-            else:
-                print '\nSelected Max Num rows (%d)' % self.rc
-        except Exception, e:
-            print e
-            import traceback
-            traceback.print_exc(file=sys.stdout)
-        self.sqlBuffer.append(self.query)
-
+            rowlimit = int(arg.parsed.suffix or 0)
+        except ValueError:
+            rowlimit = 0
+            print "Specify desired number of rows after terminator (not '%s')" % arg.parsed.suffix
+        self.varsUsed = findBinds(arg, self.binds, bindVarsIn)
+        arg = self.expandSelect(arg)
+        self.curs.execute('select ' + arg, self.varsUsed)
+        self.rows = self.curs.fetchmany(min(self.maxfetch, (rowlimit or self.maxfetch)))
+        self.rc = self.curs.rowcount
+        if self.rc > 0:
+            self.stdout.write('\n%s\n' % (self.output(arg.parsed.terminator, rowlimit)))
+        if self.rc == 0:
+            print '\nNo rows Selected.\n'
+        elif self.rc == 1: 
+            print '\n1 row selected.\n'
+            if self.autobind:
+                self.binds.update(dict(zip([''.join(l for l in d[0] if l.isalnum()) for d in self.curs.description], self.rows[0])))
+                for (i, val) in enumerate(self.rows[0]):
+                    varname = ''.join(letter for letter in self.curs.description[i][0] if letter.isalnum() or letter == '_')
+                    self.binds[varname] = val
+                    self.binds[str(i+1)] = val
+        elif self.rc < self.maxfetch:
+            print '\n%d rows selected.\n' % self.rc
+        else:
+            print '\nSelected Max Num rows (%d)' % self.rc
+        
+    def do_cat(self, arg):
+        return self.do_select(self.parsed('SELECT * FROM %s;' % arg, 
+                                          terminator = arg.parsed.terminator or ';', 
+                                          suffix = arg.parsed.suffix))
         
     @options([make_option('-f', '--full', action='store_true', help='get dependent objects as well')])
     def do_pull(self, arg, opts):
         """Displays source code."""
 
-        arg = self.parsed(arg).unterminated.upper()
-        object_type, owner, object_name = self.resolve(arg)
+        target = arg.upper()
+        object_type, owner, object_name = self.resolve(target)
         if not object_type:
             return
         self.stdout.write("%s %s.%s\n" % (object_type, owner, object_name))
@@ -568,90 +563,94 @@ class sqlpyPlus(sqlpython.sqlpython):
                 except cx_Oracle.DatabaseError:
                     pass
 
-    @options([make_option('-a','--all',action='store_true', help='Find in all schemas (not just my own)'),
-              make_option('-i', '--insensitive', action='store_true', help='case-insensitive search'),
+    all_users_option = make_option('-a', action='store_const', dest="scope",
+                                         default={'col':'', 'view':'user', 'schemas':'user'}, 
+                                         const={'col':', owner', 'view':'all', 'schemas':'all'},
+                                         help='Describe all objects (not just my own)')                
+    @options([all_users_option,
               make_option('-c', '--col', action='store_true', help='find column'),
               make_option('-t', '--table', action='store_true', help='find table')])                    
     def do_find(self, arg, opts):
         """Finds argument in source code or (with -c) in column definitions."""
-
-        arg = self.parsed(arg).unterminated.upper()     
+       
+        capArg = arg.upper()
         
         if opts.col:
-            sql = "owner, table_name, column_name from all_tab_columns where column_name like '%%%s%%'" % (arg)
+            sql = "SELECT table_name, column_name %s FROM %s_tab_columns where column_name like '%%%s%%';" \
+                % (opts.scope['col'], opts.scope['view'], capArg)
         elif opts.table:
-            sql = "owner, table_name from all_tables where table_name like '%%%s%%'" % (arg)
+            sql = "SELECT table_name %s from %s_tables where table_name like '%%%s%%';" \
+                % (opts.scope['col'], opts.scope['view'], capArg)
         else:
-            if opts.insensitive:
-                searchfor = "LOWER(text)"
-                arg = arg.lower()
-            else:
-                searchfor = "text"
-            sql = "* from all_source where %s like '%%%s%%'" % (searchfor, arg)
-        if not opts.all:
-            sql = '%s and owner = user' % (sql)
-        self.do_select(sql)
-
-    @options([make_option('-a','--all',action='store_true',
-                          help='Describe all objects (not just my own)')])
+            sql = "SELECT * from %s_source where UPPER(text) like '%%%s%%';" % (opts.scope['view'], capArg)
+        self.do_select(self.parsed(sql, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
+        
+    @options([all_users_option])
     def do_describe(self, arg, opts):
         "emulates SQL*Plus's DESCRIBE"
-
-        arg = self.parsed(arg).unterminated.upper()
-        if opts.all:
-            which_view = (', owner', 'all')
-        else:
-            which_view = ('', 'user')
-
-        if not arg:
-            self.do_select("""object_name, object_type%s FROM %s_objects WHERE object_type IN ('TABLE','VIEW','INDEX') ORDER BY object_name""" % which_view)
-            return
-        object_type, owner, object_name = self.resolve(arg)
+        target = arg.upper()
+        if not target:
+            return self.do_select(self.parsed("""SELECT object_name, object_type%s 
+                                                 FROM   %s_objects 
+                                                 WHERE  object_type IN ('TABLE','VIEW','INDEX') 
+                                                 ORDER BY object_name;""" % (opts.scope['col'], opts.scope['view']),
+                                              terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
+        object_type, owner, object_name = self.resolve(target)
         if not object_type:
-            self.do_select("""object_name, object_type%s FROM %s_objects
-                           WHERE object_type IN ('TABLE','VIEW','INDEX')
-                           AND   object_name LIKE '%%%s%%'
-                           ORDER BY object_name""" %
-                           (which_view[0], which_view[1], arg.upper()) )
-            return                    
+            return self.do_select(self.parsed("""SELECT object_name, object_type%s FROM %s_objects
+                                                 WHERE  object_type IN ('TABLE','VIEW','INDEX')
+                                                 AND    object_name LIKE '%%%s%%'
+                                                 ORDER BY object_name;""" %
+                                              (opts.scope['col'], opts.scope['view'], target),
+                                              terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
         self.stdout.write("%s %s.%s\n" % (object_type, owner, object_name))
         descQ = descQueries.get(object_type)
         if descQ:
             for q in descQ:
-                self.do_select(q,bindVarsIn={'object_name':object_name, 'owner':owner})
+                self.do_select(self.parsed(q, terminator=arg.parsed.terminator or ';' , suffix=arg.parsed.suffix), 
+                               bindVarsIn={'object_name':object_name, 'owner':owner})
         elif object_type == 'PACKAGE':
             packageContents = self.select_scalar_list(descQueries['PackageObjects'][0], {'package_name':object_name, 'owner':owner})
             for packageObj_name in packageContents:
                 self.stdout.write('Arguments to %s\n' % (packageObj_name))
-                self.do_select(descQueries['PackageObjArgs'][0],bindVarsIn={'package_name':object_name, 'owner':owner, 'object_name':packageObj_name})
+                sql = self.parsed(descQueries['PackageObjArgs'][0], terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix)
+                self.do_select(sql, bindVarsIn={'package_name':object_name, 'owner':owner, 'object_name':packageObj_name})
     do_desc = do_describe
 
     def do_deps(self, arg):
-        arg = self.parsed(arg).unterminated.upper()        
-        object_type, owner, object_name = self.resolve(arg)
+        target = arg.upper()
+        object_type, owner, object_name = self.resolve(target)
         if object_type == 'PACKAGE BODY':
             q = "and (type != 'PACKAGE BODY' or name != :object_name)'"
             object_type = 'PACKAGE'
         else:
             q = ""
-        q = """         name,
+        q = """SELECT name,
           type
           from user_dependencies
           where
           referenced_name like :object_name
           and	referenced_type like :object_type
           and	referenced_owner like :owner
-          %s""" % (q)
-        self.do_select(q, {'object_name':object_name, 'object_type':object_type, 'owner':owner})
+          %s;""" % (q)
+        self.do_select(self.parsed(q, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix), 
+                       bindVarsIn={'object_name':object_name, 'object_type':object_type, 'owner':owner})
 
     def do_comments(self, arg):
         'Prints comments on a table and its columns.'
-        arg = self.parsed(arg).unterminated.upper()        
-        object_type, owner, object_name = self.resolve(arg)
+        target = arg.upper()        
+        object_type, owner, object_name, colName = self.resolve_with_column(target)
         if object_type:
-            self.curs.execute(queries['tabComments'],{'table_name':object_name, 'owner':owner})
+            self.curs.execute(queries['tabComments'], {'table_name':object_name, 'owner':owner})
             self.stdout.write("%s %s.%s: %s\n" % (object_type, owner, object_name, self.curs.fetchone()[0]))
-            self.do_select(queries['colComments'],bindVarsIn={'owner':owner, 'object_name': object_name})
+            if colName:
+                sql = queries['oneColComments']
+                bindVarsIn={'owner':owner, 'object_name': object_name, 'column_name': colName}
+            else:
+                sql = queries['colComments'] 
+                bindVarsIn={'owner':owner, 'object_name': object_name}
+            self.do_select(self.parsed(sql, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix), 
+                           bindVarsIn=bindVarsIn)
 
     def resolve(self, identifier):
         """Checks (my objects).name, (my synonyms).name, (public synonyms).name
@@ -661,20 +660,31 @@ class sqlpyPlus(sqlpython.sqlpython):
             if len(parts) == 2:
                 owner, object_name = parts
                 object_type = self.select_scalar_list('SELECT object_type FROM all_objects WHERE owner = :owner AND object_name = :object_name',
-                                  {'owner': owner, 'object_name': object_name}
+                                  {'owner': owner, 'object_name': object_name.upper()}
                                   )[0]
             elif len(parts) == 1:
                 object_name = parts[0]
-                self.curs.execute(queries['resolve'], {'objName':object_name})
+                self.curs.execute(queries['resolve'], {'objName':object_name.upper()})
                 object_type, object_name, owner = self.curs.fetchone()
-        except TypeError:
+        except (TypeError, IndexError):
             print 'Could not resolve object %s.' % identifier
             object_type, owner, object_name = '', '', ''
         return object_type, owner, object_name
 
+    def resolve_with_column(self, identifier):
+        colName = None
+        object_type, owner, object_name = self.resolve(identifier)
+        if not object_type:
+            parts = identifier.split('.')
+            if len(parts) > 1:
+                colName = parts[-1]
+                identifier = '.'.join(parts[:-1])
+                object_type, owner, object_name = self.resolve(identifier)
+        return object_type, owner, object_name, colName
+        
     def do_resolve(self, arg):
-        arg = self.parsed(arg).unterminated.upper()        
-        self.stdout.write(','.join(self.resolve(arg))+'\n')
+        target = arg.upper()
+        self.stdout.write(','.join(self.resolve(target))+'\n')
 
     def spoolstop(self):
         if self.spoolFile:
@@ -706,6 +716,7 @@ class sqlpyPlus(sqlpython.sqlpython):
         Sorting is recommended to avoid false hits.
         Will attempt to use a graphical diff/merge tool like kdiff3, meld, or Araxis Merge, 
         if they are installed."""
+        #TODO: Update this to use pyparsing
         fnames = []
         args2 = args.split(' to ')
         if len(args2) < 2:
@@ -757,59 +768,40 @@ class sqlpyPlus(sqlpython.sqlpython):
         except IndexError:
             args = ''
         try:
-            return self.onecmd('%s %s' % (commands[abbrev], args))
+            return self.onecmd('%s %s%s%s' % (commands[abbrev], args, arg.parsed.terminator, arg.parsed.suffix))
         except KeyError:
             print 'psql command \%s not yet supported.' % abbrev
 
-    @options([make_option('-a','--all',action='store_true',
-                          help='Describe all objects (not just my own)')])
+    @options([all_users_option])
     def do__dir_tables(self, arg, opts):
-        if opts.all:
-            which_view = (', owner', 'all')
-        else:
-            which_view = ('', 'user')        
-        self.do_select("""table_name, 'TABLE' as type%s FROM %s_tables WHERE table_name LIKE '%%%s%%'""" %
-                       (which_view[0], which_view[1], arg.upper()))        
-
-    @options([make_option('-a','--all',action='store_true',
-                          help='Describe all objects (not just my own)')])
+        sql = """SELECT table_name, 'TABLE' as type%s FROM %s_tables WHERE table_name LIKE '%%%s%%';""" % \
+                       (opts.scope['col'], opts.scope['view'], arg.upper())
+        self.do_select(self.parsed(sql, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
+        
+    @options([all_users_option])
     def do__dir_views(self, arg, opts):
-        if opts.all:
-            which_view = (', owner', 'all')
-        else:
-            which_view = ('', 'user')        
-        self.do_select("""view_name, 'VIEW' as type%s FROM %s_views WHERE view_name LIKE '%%%s%%'""" %
-                       (which_view[0], which_view[1], arg.upper())) 
-
-    @options([make_option('-a','--all',action='store_true',
-                          help='Describe all objects (not just my own)')])
+        sql = """SELECT view_name, 'VIEW' as type%s FROM %s_views WHERE view_name LIKE '%%%s%%';""" % \
+                       (opts.scope['col'], opts.scope['view'], arg.upper())
+        self.do_select(self.parsed(sql, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
+        
+    @options([all_users_option])
     def do__dir_indexes(self, arg, opts):
-        if opts.all:
-            which_view = (', owner', 'all')
-        else:
-            which_view = ('', 'user')        
-        self.do_select("""index_name, index_type%s FROM %s_indexes WHERE index_name LIKE '%%%s%%' OR table_name LIKE '%%%s%%'""" %
-                       (which_view[0], which_view[1], arg.upper(), arg.upper())) 
+        sql = """SELECT index_name, index_type%s FROM %s_indexes WHERE index_name LIKE '%%%s%%' OR table_name LIKE '%%%s%%';""" % \
+                       (opts.scope['col'], opts.scope['view'], arg.upper(), arg.upper())
+        self.do_select(self.parsed(sql, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
 
     def do__dir_tablespaces(self, arg):
-        self.do_select("""tablespace_name, file_name from dba_data_files""") 
+        sql = """SELECT tablespace_name, file_name from dba_data_files;"""
+        self.do_select(self.parsed(sql, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
 
     def do__dir_schemas(self, arg):
-        self.do_select("""owner, count(*) AS objects FROM all_objects GROUP BY owner ORDER BY owner""") 
+        sql = """SELECT owner, count(*) AS objects FROM all_objects GROUP BY owner ORDER BY owner;"""
+        self.do_select(self.parsed(sql, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
 
     def do_head(self, arg):
-        nrows = 10
-        args = arg.split()
-        if len(args) > 1:
-            for a in args:
-                if a[0] == '-':
-                    try:
-                        nrows = int(a[1:])
-                        args.remove(a)
-                    except:
-                        pass
-            arg = ' '.join(args)
-        self.do_select('* from %s;%d' % (arg, nrows))
+        sql = self.parsed('SELECT * FROM %s;' % arg, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix)
+        sql.parsed['suffix'] = sql.parsed.suffix or '10'
+        self.do_select(self.parsed(sql))
 
     def do_print(self, arg):
         'print VARNAME: Show current value of bind variable VARNAME.'
@@ -826,7 +818,8 @@ class sqlpyPlus(sqlpython.sqlpython):
 
     assignmentScanner = Parser(pyparsing.Literal(':=') ^ '=')
     def do_setbind(self, arg):
-        arg = self.parsed(arg).unterminated
+        if not arg:
+            return self.do_print(arg)
         try:
             assigner, startat, endat = self.assignmentScanner.scanner.scanString(arg).next()
         except StopIteration:
@@ -846,14 +839,13 @@ class sqlpyPlus(sqlpython.sqlpython):
             except ValueError: 
                 statekeeper = Statekeeper(self, ('autobind',))  
                 self.autobind = True
-                self.do_select('%s AS %s FROM dual;' % (val, var))
+                self.onecmd('SELECT %s AS %s FROM dual;' % (val, var))
                 statekeeper.restore()
 
     def do_exec(self, arg):
         if arg[0] == ':':
             self.do_setbind(arg[1:])
         else:
-            arg = self.parsed(arg).unterminated
             varsUsed = findBinds(arg, self.binds, {})
             try:
                 self.curs.execute('begin\n%s;end;' % arg, varsUsed)
@@ -882,9 +874,6 @@ class sqlpyPlus(sqlpython.sqlpython):
     def do_declare(self, arg):
         self.anon_plsql('declare ' + arg)
 
-    #def do_create(self, arg):
-    #    self.anon_plsql('create ' + arg)
-
     @options([make_option('-l', '--long', action='store_true', help='long descriptions'),
               make_option('-a', '--all', action='store_true', help="all schemas' objects")])        
     def do_ls(self, arg, opts):
@@ -895,44 +884,25 @@ class sqlpyPlus(sqlpython.sqlpython):
         else:
             where = ''
         if opts.all:
-            owner = 'owner'
             whose = 'all'
+            objname = "owner || '.' || object_name"            
         else:
-            owner = "'' AS owner"
             whose = 'user'
-        result = []
-        statement = '''SELECT object_type, object_name,
-                  status, last_ddl_time, %s
+            objname = 'object_name'            
+        if opts.long:
+            extraInfo = ', status, last_ddl_time AS modified'
+        else:
+            extraInfo = ''
+        statement = '''SELECT object_type || '/' || %s AS name %s 
                   FROM   %s_objects %s
-                  ORDER BY object_type, object_name''' % (owner, whose, where)
-        self.curs.execute(statement)
-        for (object_type, object_name, status, last_ddl_time, owner) in self.curs.fetchall():
-            if opts.all:
-                qualified_name = '%s.%s' % (owner, object_name)
-            else:
-                qualified_name = object_name
-            if opts.long:
-                result.append('%s\t%s\t%s/%s' % (status, last_ddl_time, object_type, qualified_name))
-            else:
-                result.append('%s/%s' % (object_type, qualified_name))
-        self.stdout.write('\n'.join(result) + '\n')
-
-    def do_cat(self, arg):
-        '''cat TABLENAME --> SELECT * FROM equivalent'''
-        if not arg:
-            print self.do_cat.__doc__
-            return
-        arg = self.parsed(arg)
-        targets = arg.unterminated.split()
-        for target in targets:
-            self.do_select('* from %s%s%s' % (target, arg.terminator, arg.rowlimit)) # permissive of space before terminator
-
+                  ORDER BY object_type, object_name;''' % (objname, extraInfo, whose, where)
+        self.do_select(self.parsed(statement, terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix))
+        
     @options([make_option('-i', '--ignore-case', dest='ignorecase', action='store_true', help='Case-insensitive search')])        
     def do_grep(self, arg, opts):
         """grep PATTERN TABLE - search for term in any of TABLE's fields"""    
 
-        arg = self.parsed(arg)
-        targetnames = arg.unterminated.split()
+        targetnames = arg.split()
         pattern = targetnames.pop(0)
         targets = [] 
         for target in targetnames:
@@ -946,30 +916,71 @@ class sqlpyPlus(sqlpython.sqlpython):
         for target in targets:
             print target
             target = target.rstrip(';')
-            sql = []
             try:
                 self.curs.execute('select * from %s where 1=0' % target) # just to fill description
                 if opts.ignorecase:
                     sql = ' or '.join("LOWER(%s) LIKE '%%%s%%'" % (d[0], pattern.lower()) for d in self.curs.description)                                        
                 else:
                     sql = ' or '.join("%s LIKE '%%%s%%'" % (d[0], pattern) for d in self.curs.description)
-                sql = '* FROM %s WHERE %s' % (target, sql)
-                self.do_select('%s%s%s' % (sql, arg.terminator, arg.rowlimit))
+                sql = self.parsed('SELECT * FROM %s WHERE %s;' % (target, sql), terminator=arg.parsed.terminator or ';', suffix=arg.parsed.suffix)
+                self.do_select(sql)
             except Exception, e:
                 print e
                 import traceback
                 traceback.print_exc(file=sys.stdout)                
 
     def do_refs(self, arg):
-        arg = self.parsed(arg).unterminated.upper()        
-        object_type, owner, object_name = self.resolve(arg)
-        if object_type == 'TABLE':
-            self.do_select(queries['refs'],bindVarsIn={'object_name':object_name, 'owner':owner})
+        result = []
+        (type, owner, table_name) = self.resolve(arg.upper())        
+        self.curs.execute("""SELECT constraint_name, r_owner, r_constraint_name 
+                             FROM   all_constraints 
+                             WHERE  constraint_type = 'R'
+                             AND    owner = :owner
+                             AND    table_name = :table_name""",
+                             {"owner": owner, "table_name": table_name})
+        for (constraint_name, remote_owner, remote_constraint_name) in self.curs.fetchall():
+            result.append('%s on %s.%s:' % (constraint_name, owner, table_name))
+            self.curs.execute("SELECT column_name FROM all_cons_columns WHERE owner = :owner AND constraint_name = :constraint_name ORDER BY position",
+                              {'constraint_name': constraint_name, 'owner': owner})
+            result.append("    (%s)" % (",".join(col[0] for col in self.curs.fetchall())))
+            self.curs.execute("SELECT table_name FROM all_constraints WHERE owner = :remote_owner AND constraint_name = :remote_constraint_name",
+                              {'remote_owner': remote_owner, 'remote_constraint_name': remote_constraint_name})
+            remote_table_name = self.curs.fetchone()[0]
+            result.append("must be in %s:" % (remote_table_name))
+            self.curs.execute("SELECT column_name FROM all_cons_columns WHERE owner = :remote_owner AND constraint_name = :remote_constraint_name ORDER BY position",
+                              {'remote_constraint_name': remote_constraint_name, 'remote_owner': remote_owner})
+            result.append('    (%s)\n' % (",".join(col[0] for col in self.curs.fetchall())))
+        remote_table_name = table_name
+        remote_owner = owner
+        self.curs.execute("""SELECT  owner, constraint_name, table_name, r_constraint_name
+                             FROM    all_constraints
+                             WHERE   (r_owner, r_constraint_name) IN
+                               ( SELECT owner, constraint_name
+                                 FROM   all_constraints
+                                 WHERE  table_name = :remote_table_name
+                                 AND    owner = :remote_owner )""",
+                          {'remote_table_name': remote_table_name, 'remote_owner': remote_owner})
+        for (owner, constraint_name, table_name, remote_constraint_name) in self.curs.fetchall():
+            result.append('%s on %s.%s:' % (constraint_name, owner, table_name))
+            self.curs.execute("SELECT column_name FROM all_cons_columns WHERE owner = :owner AND constraint_name = :constraint_name ORDER BY position",
+                              {'constraint_name': constraint_name, 'owner': owner})
+            result.append("    (%s)" % (",".join(col[0] for col in self.curs.fetchall())))
+            self.curs.execute("SELECT table_name FROM all_constraints WHERE owner = :remote_owner AND constraint_name = :remote_constraint_name",
+                              {'remote_owner': remote_owner, 'remote_constraint_name': remote_constraint_name})
+            remote_table_name = self.curs.fetchone()[0]
+            result.append("must be in %s:" % (remote_table_name))
+            self.curs.execute("SELECT column_name FROM all_cons_columns WHERE owner = :remote_owner AND constraint_name = :remote_constraint_name ORDER BY position",
+                              {'remote_constraint_name': remote_constraint_name, 'remote_owner': remote_owner})
+            result.append('    (%s)\n' % (",".join(col[0] for col in self.curs.fetchall())))
+        self.stdout.write('\n'.join(result) + "\n")
     
 def _test():
     import doctest
     doctest.testmod()
-
+    
+if __name__ == "__main__":
+    "Silent return implies that all unit tests succeeded.  Use -v to see details."
+    _test()
 if __name__ == "__main__":
     "Silent return implies that all unit tests succeeded.  Use -v to see details."
     _test()
